var assert = require('assert');
var debug = require('debug')('docker-exec-websocket-client:lib:client');
var debugdata = require('debug')('docker-exec-websocket-client:lib:data');
var EventEmitter = require('events').EventEmitter;
var msgcode = require('./messagecodes.js');
var querystring = require('querystring');
var through2 = require('through2').obj;
var WS = require('ws');

var BROWSER = typeof window === 'undefined';

class DockerExecWebsocketClient extends EventEmitter {
  constructor(options) {
    super();
    this.options = {
      tty: true,
      command: 'sh',
      wsopts: {},
      ...options,
    };
  }

  /* Makes a client program with unbroken stdin, stdout, stderr streams
   * Is also an EventEmitter with 'exit' event
   *
   * Required options:
   * url parts (hostname port pathname)
   * or url
   * tty: whether or not we expect VT100 style output
   * command: array or string of command to be run in exec
   */
  async execute() {
    this.url = this.options.url + '?' + querystring.stringify({
      tty: this.options.tty ? 'true' : 'false',
      command: this.options.command,
    });
    debug(this.url);
    assert(/ws?s:\/\//.test(this.url), 'url required or malformed url input');

    //HACK: browser check
    if (BROWSER) { //means that this is probably node
      this.socket = new WS(this.url, this.options.wsopts);
    } else { //means this is probably a browser, which means we ignore options
      this.socket = new WebSocket(this.url);
    }

    this.socket.binaryType = 'arraybuffer';
    this.socket.addEventListener('open', () => {
      debug('socket opened');
      this.emit('open');
    });

    this.stdin = through2((data, enc, cb) => {
      debug('through2.stdin.1: sendMessage %s' % data);
      this.sendMessage(msgcode.stdin, data);
      debug('through2.stdin.1: callback');
      cb();
      debug('through2.stdin.1: done');
    }, (cb) => {
      debug('through2.stdin.2: sendCode end');
      this.sendCode(msgcode.end);
      debug('through2.stdin.2: callback');
      cb();
      debug('through2.stdin.2: done');
    });

    const MAX_OUTSTANDING_BYTES = 8 * 1024 * 1024;
    this.outstandingBytes = 0;

    //stream with pause buffering, everything passes thru here first
    this.strbuf = through2();
    this.strbuf.on('data', (data) => {
      this.outstandingBytes += data.length;
      debug('received %s bytes (%s) now %s to send', data.length, data, this.outstandingBytes);
      if (BROWSER) {
        this.socket.send(data, {binary: true}, () => {
          this.outstandingBytes -= data.length;
          debug('sent %s bytes (BROWSER, binary), now %s to send', data.length, this.outstandingBytes);
        });
      } else {
        this.socket.send(data);
        this.outstandingBytes -= data.length;
        debug('sent %s bytes (non-BROWSER), now %s to send', data.length, this.outstandingBytes);
      }
      if (this.outstandingBytes > MAX_OUTSTANDING_BYTES) {
        this.strbuf.pause();
        this.emit('paused');
        debug('paused');
      } else {
        this.strbuf.resume();
        this.emit('resumed');
        debug('resumed');
      }
    });
    //Starts out paused so that input isn't sent until server is ready
    this.strbuf.pause();

    this.stdout = through2();
    this.stderr = through2();
    this.stdout.draining = false;
    this.stderr.draining = false;

    this.stdout.on('drain', () => {
      debug('stdout drain, stdout / stderr draining: %s / %s', this.stdout.draining, this.stderr.draining);
      this.stdout.draining = false;
      if (!this.stderr.draining) {
        this.sendCode(msgcode.resume);
      }
    });

    this.stderr.on('drain', () => {
      debug('stderr drain, stdout / stderr draining: %s / %s', this.stdout.draining, this.stderr.draining);
      this.stderr.draining = false;
      if (!this.stdout.draining) {
        this.sendCode(msgcode.resume);
      }
    });

    this.socket.onmessage = (messageEvent) => {
      this.messageHandler(messageEvent);
    };
    await new Promise((accept, reject) => {
      this.socket.addEventListener('error', reject);
      this.socket.addEventListener('open', accept);
    });
    this.socket.addEventListener('error', err => this.emit('error', err));
    debug('client connected');
  }

  messageHandler(messageEvent) {
    var message = Buffer.from(new Uint8Array(messageEvent.data));
    debugdata(message);
    // the first byte is the message code
    switch (message[0]) {
      //pauses the client, causing strbuf to buffer
      case msgcode.pause:
        debug('message:paused');
        this.strbuf.pause();
        this.emit('paused');
        break;

      //resumes the client, flushing strbuf
      case msgcode.resume:
        debug('message:resumed');
        this.strbuf.resume();
        this.emit('resumed');
        break;

      case msgcode.stdout:
        debug('message:stdout');
        if (!this.stdout.write(message.slice(1))) {
          this.sendCode(msgcode.pause);
          this.stdout.draining = true;
        }
        break;

      case msgcode.stderr:
        debug('message:stderr');
        if (!this.stderr.write(message.slice(1))) {
          this.sendCode(msgcode.pause);
          this.stderr.draining = true;
        }
        break;

      //first byte contains exit code
      case msgcode.stopped:
        debug('message:stopped');
        this.emit('exit', message.readInt8(1));
        this.close();
        break;

      case msgcode.shutdown:
        debug('message:shutdown');
        this.emit('shutdown');
        this.close();
        break;

      case msgcode.error:
        debug('message:error');
        this.emit('error', message.slice(1));
        break;

      default:
        debug('message:unknown msg code %s', message[0]);
    }
  }

  resize(h, w) {
    if (!this.options.tty) {
      throw new Error('cannot resize, not a tty instance');
    } else {
      var buf = Buffer.alloc(4);
      buf.writeUInt16LE(h, 0);
      buf.writeUInt16LE(w, 2);
      debug('resized to %sx%s', h, w);
      this.sendMessage(msgcode.resize, buf);
    }
  }

  sendCode(code) {
    this.strbuf.write(Buffer.from([code]));
  }

  sendMessage(code, data) {
    this.strbuf.write(Buffer.concat([Buffer.from([code]), Buffer.from(data)]));
  }

  close() {
    debug("close()")
    if (!this.strbuf.paused) {
      debug("close(): socket");
      this.socket.close();
      debug("close(): stdin");
      this.stdin.end();
      debug("close(): stdout");
      this.stdout.end();
      debug("close(): stderr");
      this.stderr.end();
      debug("close(): stdbuf");
      this.strbuf.end();
    } else {
      debug("close(): waiting to drain...")
      this.strbuf.on('drain', () => {
        debug("close() after drain: socket");
        this.socket.close();
        debug("close() after drain: stdin");
        this.stdin.end();
        debug("close() after drain: stdout");
        this.stdout.end();
        debug("close() after drain: stderr");
        this.stderr.end();
        debug("close() after drain: stdbuf");
        this.strbuf.end();
      });
    }
  }
}

module.exports = DockerExecWebsocketClient;
